"""
This module implements hooks to improve the packaging process for Python
projects that use Git.
"""

import os
from subprocess import Popen, PIPE, STDOUT


def cmd_out(cmd):
    """
    Helper for running shell commands.
    """
    wd = os.path.dirname(os.path.realpath(__file__))
    p = Popen(cmd, shell=True, cwd=wd, stdout=PIPE, stderr=STDOUT)
    r, _ = p.communicate()
    res = r.strip().decode()
    return res, p.returncode


def get_version():
    """
    Get the version, first by trying to read the ``VERSION_STAMP``, and if that
    fails, by attempting to use ``git`` annotated tags.
    """
    git_options = ['git', '/usr/local/bin/git', '/usr/bin/git']
    d = os.path.dirname(os.path.realpath(__file__))
    try:
        x = open(os.path.join(d, 'VERSION_STAMP'), 'rb').read().strip().decode()
        if x: return x
    except OSError:
        pass
    for g in git_options:
        gitver, _ = cmd_out('{0} describe --tags --always'.format(g))
        gitchanged, _ = cmd_out('{0} diff-index HEAD --'.format(g))
        gitver = gitver.replace('v', '').split('-g')[0].replace('-', '.')
        if ' ' in gitver: gitver = '0'
        else: break
    if not gitver or 'fatal' in gitver or '\n' in gitver:
        return '0'
    if gitchanged.strip():
        gitver += '+changed'
    return gitver


def stamp_directory(d):
    """
    Write the ``VERSION_STAMP`` file.
    """
    v = get_version()
    with open(os.path.join(d, 'VERSION_STAMP'), 'wb') as f:
        f.write(v.encode() + b'\n')


def unstamp_directory(d):
    """
    Remove the ``VERSION_STAMP`` file.
    """
    os.remove(os.path.join(d, 'VERSION_STAMP'))


if __name__ == '__main__':
    print(get_version())



"""
This module is designed to do 2 things:

 1. Provide an interface (``get_setting``) for getting settings from
    ``django.conf.settings`` with settings in this file being the fallback
    default.
 2. Provide enough settings for this app to run as a standalone project.
"""

from django.conf import settings
import os
import sys


me = sys.modules[__name__]


def get_setting(name):
    """
    Hook for getting Django settings and using properties of this file as the
    default.
    """
    try:
        return getattr(settings, name, getattr(me, name, None))
    except NameError:
        pass
    return None


# setting `base_dir` before defaults since it needs to be used by `DATABASES`.
base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# standalone settings
BASE_DIR = base_dir
SECRET_KEY = 'not-a-very-good-secret'
DEBUG = True
ALLOWED_HOSTS = ['*']
INSTALLED_APPS = [
    'action_serializer',
    'django.contrib.admin',
    'django.contrib.admindocs',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework',
    'rest_framework.authtoken',
]
MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]
ROOT_URLCONF = 'action_serializer.urls'
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]
WSGI_APPLICATION = 'action_serializer.wsgi.application'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(base_dir, 'db.sqlite3'),
    }
}
AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]
REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework.authentication.TokenAuthentication',
    ),
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.LimitOffsetPagination',
    'PAGE_SIZE': 10,
}
LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'America/Chicago'
USE_I18N = True
USE_L10N = True
USE_TZ = True
STATIC_URL = '/static/'
APPEND_SLASH = True

# local development settings
try:
    from .local_settings import *
except ImportError:
    pass



"""
Tests for the drf-action-serializer package.
"""
# pylint: disable=unused-variable

import os

from django.db import models
from django.test import TestCase

from . import version
from .serializers import ModelActionSerializer


def dedent(blocktext):
    return '\n'.join([line[12:] for line in blocktext.splitlines()[1:-1]])


class RegularFieldsModel(models.Model):
    """
    A model class for testing regular flat fields.
    """
    auto_field = models.AutoField(primary_key=True)
    boolean_field = models.BooleanField(default=False)
    char_field = models.CharField(max_length=100)
    date_field = models.DateField()
    decimal_field = models.DecimalField(max_digits=3, decimal_places=1)
    email_field = models.EmailField(max_length=100)
    float_field = models.FloatField()
    integer_field = models.IntegerField()

    def method(self):
        return 'method'


class ModelActionSerializerTestCase(TestCase):
    """
    Tests for the ModelActionSerializer.
    """

    def test_action_fields(self):
        """
        Ensure `action_fields` are used if the context declares a list view.
        """
        class TestSerializer(ModelActionSerializer):
            class Meta:
                model = RegularFieldsModel
                fields = ('char_field')
                action_fields = {
                    'list': {'fields': ('auto_field', 'char_field')}
                }

        expected = dedent("""
            TestSerializer(context={'view': <class 'action_serializer.tests.ActionView'>}):
                auto_field = IntegerField(read_only=True)
                char_field = CharField(max_length=100)
        """)
        context = {'view': type('ActionView', (object,), {'action': 'list'})}
        self.assertEqual(repr(TestSerializer(context=context)), expected)

    def test_action_fields_different_action(self):
        """
        Ensure normal `fields` are used if the context view action is different
        than what is defined in the `action_fields`
        """
        class TestSerializer(ModelActionSerializer):
            class Meta:
                model = RegularFieldsModel
                fields = ('char_field',)
                action_fields = {
                    'list': {'fields': ('auto_field', 'char_field')}
                }

        expected = dedent("""
            TestSerializer(context={'view': <class 'action_serializer.tests.ActionView'>}):
                char_field = CharField(max_length=100)
        """)
        context = {'view': type('ActionView', (object,), {'action': 'create'})}
        self.assertEqual(repr(TestSerializer(context=context)), expected)

    def test_action_extra_kwargs(self):
        """
        Ensure `action_fields` `extra_kwargs` are used if they are defined and
        the context view action matches.
        """
        class TestSerializer(ModelActionSerializer):
            class Meta:
                model = RegularFieldsModel
                fields = ('char_field')
                action_fields = {
                    'list': {
                        'fields': ('auto_field', 'char_field'),
                        'extra_kwargs': {
                            'auto_field': {
                                'required': False,
                                'read_only': False,
                            },
                        }
                    }
                }

        expected = dedent("""
            TestSerializer(context={'view': <class 'action_serializer.tests.ActionView'>}):
                auto_field = IntegerField(read_only=False, required=False)
                char_field = CharField(max_length=100)
        """)
        context = {'view': type('ActionView', (object,), {'action': 'list'})}
        self.assertEqual(repr(TestSerializer(context=context)), expected)

    def test_action_exclude(self):
        """
        Ensure `action_fields` `exclude` is used if it is defined and the
        context view action matches.
        """
        auto_field = models.AutoField(primary_key=True)
        boolean_field = models.BooleanField(default=False)
        char_field = models.CharField(max_length=100)
        date_field = models.DateField()
        decimal_field = models.DecimalField(max_digits=3, decimal_places=1)
        email_field = models.EmailField(max_length=100)
        float_field = models.FloatField()
        integer_field = models.IntegerField()
        class TestSerializer(ModelActionSerializer):
            class Meta:
                model = RegularFieldsModel
                fields = ('char_field')
                action_fields = {'list': {'exclude': (
                    'boolean_field',
                    'date_field',
                    'float_field',
                    'integer_field',
                )}}

        expected = dedent("""
            TestSerializer(context={'view': <class 'action_serializer.tests.ActionView'>}):
                auto_field = IntegerField(read_only=True)
                char_field = CharField(max_length=100)
                decimal_field = DecimalField(decimal_places=1, max_digits=3)
                email_field = EmailField(max_length=100)
        """)
        context = {'view': type('ActionView', (object,), {'action': 'list'})}
        self.assertEqual(repr(TestSerializer(context=context)), expected)


class VersionTestCase(TestCase):
    """
    Tests for versioning script.
    """

    def setUp(self):
        # put us into the package directory
        os.chdir(os.path.normpath(os.path.dirname(os.path.abspath(__file__))))

    def test_version_is_str(self):
        self.assertTrue(isinstance(version.get_version(), str))

    def test_stamp_unstamp(self):
        stamp_location = './VERSION_STAMP'
        try:
            os.remove(stamp_location)
        except OSError:
            pass
        version.stamp_directory('.')
        self.assertTrue(os.path.exists(stamp_location))
        self.assertTrue(os.path.isfile(stamp_location))
        version.unstamp_directory('.')
        self.assertFalse(os.path.exists(stamp_location))



from django.contrib import admin
from django.urls import include, path
from rest_framework.routers import DefaultRouter

from .sample_group_viewset import GroupViewSet


router = DefaultRouter()
router.register('auth/group', GroupViewSet)
router.register('auth/groups', GroupViewSet)

urlpatterns = [
    path('api/', include(router.urls)),
    path('admin/doc/', include('django.contrib.admindocs.urls')),
    path('admin/', admin.site.urls),
]



"""
This module implements the ``ModelActionSerializer`` which is a
``ModelSerializer`` but which accepts more ``Meta`` attributes for per-action
configurations of the fields.
"""

from rest_framework.serializers import *


class ModelActionSerializer(ModelSerializer):
    """
    Similar to ModelSerializer, except now in `get_field_names` and
    `get_extra_kwargs`, if a key for the current action exists in the
    `Meta.action_fields` dictionary, then abide by the `fields`, `exclude`, and
    `extra_kwargs` keys of the resulting dictionary.
    """

    def get_action_config(self):
        """
        Return the configuration in the `Meta.action_fields` dictionary for this
        view's action.
        """
        view = getattr(self, 'context', {}).get('view', None)
        action = getattr(view, 'action', None)
        action_fields = getattr(self.Meta, 'action_fields', {})
        return action_fields.get(action, None)

    def get_field_names(self, declared_fields, info):
        """
        Returns the list of all field names that should be created when
        instantiating this serializer class. This is based on the default
        set of fields, but also takes into account the `Meta.fields` or
        `Meta.exclude` options if they have been specified, and also the
        `fields` and `exclude` properties for the action in `Meta.action_fields`
        dictionary.
        """

        action_config = self.get_action_config()
        if action_config:
            fields = action_config.get('fields', None)
            exclude = action_config.get('exclude', None)
        else:
            fields = getattr(self.Meta, 'fields', None)
            exclude = getattr(self.Meta, 'exclude', None)

        if fields and fields != ALL_FIELDS and not isinstance(fields, (list, tuple)):
            raise TypeError(
                'The `fields` option must be a list or tuple or "__all__". '
                'Got %s.' % type(fields).__name__
            )

        if exclude and not isinstance(exclude, (list, tuple)):
            raise TypeError(
                'The `exclude` option must be a list or tuple. Got %s.' %
                type(exclude).__name__
            )

        assert not (fields and exclude), (
            "Cannot set both 'fields' and 'exclude' options on "
            "serializer {serializer_class}.".format(
                serializer_class=self.__class__.__name__
            )
        )

        assert not (fields is None and exclude is None), (
            "Creating a ModelSerializer without either the 'fields' attribute "
            "or the 'exclude' attribute has been deprecated since 3.3.0, "
            "and is now disallowed. Add an explicit fields = '__all__' to the "
            "{serializer_class} serializer.".format(
                serializer_class=self.__class__.__name__
            ),
        )

        if fields == ALL_FIELDS:
            fields = None

        if fields is not None:
            # Ensure that all declared fields have also been included in the
            # `Meta.fields` option.

            # Do not require any fields that are declared in a parent class,
            # in order to allow serializer subclasses to only include
            # a subset of fields.
            required_field_names = set(declared_fields)
            for cls in self.__class__.__bases__:
                required_field_names -= set(getattr(cls, '_declared_fields', []))

            for field_name in required_field_names:
                assert field_name in fields, (
                    "The field '{field_name}' was declared on serializer "
                    "{serializer_class}, but has not been included in the "
                    "'fields' option.".format(
                        field_name=field_name,
                        serializer_class=self.__class__.__name__
                    )
                )
            return fields

        # Use the default set of field names if `Meta.fields` is not specified.
        fields = self.get_default_field_names(declared_fields, info)

        if exclude is not None:
            # If `Meta.exclude` is included, then remove those fields.
            for field_name in exclude:
                assert field_name not in self._declared_fields, (
                    "Cannot both declare the field '{field_name}' and include "
                    "it in the {serializer_class} 'exclude' option. Remove the "
                    "field or, if inherited from a parent serializer, disable "
                    "with `{field_name} = None`."
                    .format(
                        field_name=field_name,
                        serializer_class=self.__class__.__name__
                    )
                )

                assert field_name in fields, (
                    "The field '{field_name}' was included on serializer "
                    "{serializer_class} in the 'exclude' option, but does "
                    "not match any model field.".format(
                        field_name=field_name,
                        serializer_class=self.__class__.__name__
                    )
                )
                fields.remove(field_name)

        return fields

    def get_extra_kwargs(self):
        """
        Return a dictionary mapping field names to a dictionary of
        additional keyword arguments.
        """
        action_config = self.get_action_config()
        if action_config:
            extra_kwargs = copy.deepcopy(action_config.get('extra_kwargs', {}))
        else:
            extra_kwargs = copy.deepcopy(getattr(self.Meta, 'extra_kwargs', {}))

        read_only_fields = getattr(self.Meta, 'read_only_fields', None)
        if read_only_fields is not None:
            if not isinstance(read_only_fields, (list, tuple)):
                raise TypeError(
                    'The `read_only_fields` option must be a list or tuple. '
                    'Got %s.' % type(read_only_fields).__name__
                )
            for field_name in read_only_fields:
                kwargs = extra_kwargs.get(field_name, {})
                kwargs['read_only'] = True
                extra_kwargs[field_name] = kwargs

        else:
            # Guard against the possible misspelling `readonly_fields` (used
            # by the Django admin and others).
            assert not hasattr(self.Meta, 'readonly_fields'), (
                'Serializer `%s.%s` has field `readonly_fields`; '
                'the correct spelling for the option is `read_only_fields`.' %
                (self.__class__.__module__, self.__class__.__name__)
            )

        return extra_kwargs



"""
This module implements an example ``ModelAdminSerializer`` and ViewSet for the
Django ``Group`` model.
"""

from django.contrib.auth.models import Group
from rest_framework.viewsets import ModelViewSet

from .serializers import ModelActionSerializer


class GroupActionSerializer(ModelActionSerializer):
    """
    An example serializer for the Django ``Group`` model with details, and the
    list view has less fields than the detail.
    """

    class Meta:
        model = Group
        fields = ('id', 'name', 'permissions')
        action_fields = {
            'list': {
                'fields': ('id', 'name'),
            },
        }


class GroupViewSet(ModelViewSet):
    """
    An example viewset for the Django ``Group`` model.
    """
    serializer_class = GroupActionSerializer
    queryset = Group.objects.all()



import os
from setuptools import find_packages, setup
from action_serializer import version


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# stamp the package prior to installation
version.stamp_directory('./action_serializer')

# get README
with open('README.rst') as f:
    long_description = f.read()

setup(
    name='drf-action-serializer',
    version=version.get_version(),
    packages=find_packages(),
    include_package_data=True,
    package_data={'action_serializer': ['VERSION_STAMP']},
    install_requires=['Django>=2', 'djangorestframework>=3'],
    description='A serializer that handles field configuration with multiple actions from ViewSets.',
    long_description=long_description,
    url='https://github.com/gregschmit/drf-action-serializer',
    author='Gregory N. Schmit',
    author_email='schmitgreg@gmail.com',
    license='MIT',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)

# un-stamp the package after installation
version.unstamp_directory('./action_serializer')
